module cpu_bfm (ADDR, ADS_N, CLKSEL, RD_N, WR_N, EN_N, INTR, DATA);

output [2:0] ADDR;
output       ADS_N;
output [7:0] CLKSEL;
output       RD_N;
output       WR_N;
output       EN_N;



inout        INTR;

inout  [7:0] DATA;

parameter div = 1;                      // division factor for timebase

// Data Bus Interface Timing
parameter tads     = 60,    //     60ns  <=     Tads
          tah      = 0,     //      0ns	 <=     Tah 
          tar      = 60,    //     60ns  <=     Tar 
          tas      = 60,    //     60ns  <=     Tas 
          taw      = 60,    //     60ns  <=     Taw 
          tdh      = 20,    //     20ns  <=     Tdh
          tds      = 30,    //     30ns  <=     Tds 
          tes      = 60,    //     60ns  <=     Tes 
          thz      = 100,   //      0ns  <=     Thz  <=  100ns
          tpdi     = 100,   //      0ns  <=     Tpdi <=  100ns
          trd      = 125,   //    125ns  <=     Trd
          tren     = 60,    //      0ns  <=     Tren <= 60ns
          trvd     = 60,    //      0ns  <=     Trvd <= 60ns
          twen     = 20,    //     20ns  <=     Twen 
          twr      = 100,   //    100ns  <=     Twr 
          twrst    = 100,   //                  Twrst  <=  100ns
          twrt     = 100;   //      0ns  <=     Twrt <=  100ns 

// OneWire Interface Timing
parameter trstl     = 501000,    //     500.8us  <=     Trstl  <= 626us
          tr        = 0,         //      0ns	 <=     Tah 
          trsth     = 510000;    //     508.8us  <=     Trsth  <= 636us     



// one wire command bytes
parameter [7:0] READ_ROM	= 8'h33,
                MATCH_ROM	= 8'h55,
                SEARCH_ROM	= 8'hF0,
		COND_SEARCH_ROM = 8'hEC,
                SKIP_ROM	= 8'hCC;

// DS1WM memory functions
parameter [7:0] WRITE_SP        = 8'h0F,
                READ_SP         = 8'hAA,
                COPY_SP         = 8'h55,
                READ_MEM        = 8'hF0;
		
// DS1WM ADRESS MAPPED REGISTERS

parameter [2:0] COMMAND_REG     = 3'd0,
                TX_BUFFER       = 3'd1,
                RX_BUFFER       = 3'd1,
                IR              = 3'd2,
                IR_ENABLE       = 3'd3,
                CLK_DIV_REG     = 3'd4,
                CNTL_REG        = 3'd5;

reg ias;
reg [7:0] ir_en_reg;

reg [7:0] current_cmd ;
reg [7:0] dout;

reg [2:0] ADDR;
reg       ADS_N;
reg [7:0] CLKSEL;
reg       RD_N;
reg       WR_N;
reg       EN_N;

//reg [7:0] ir_reg;
wire       INTR;
wire       intr_int = INTR ^ ias;

wire [7:0] DATA;


assign DATA = dout;
//-----------------------------------------------------------

initial begin
   ADDR = 3'd0;
   ADS_N = 1'b1;
   RD_N  = 1'b1;
   WR_N  = 1'b1;
   EN_N  = 1'b1;
   
   dout = 8'hz;
   ias = 0;
end


task ow_reset;            // 1-Wire Reset ////////////////////////////////////////

	
begin

  ir_en_reg = 8'h01 ^ {6'b000000,ias,1'b0};
  
  cpu_write(IR_ENABLE,ir_en_reg);
  
  ADS_N = 1'b0;
  
  # (tads - tas);
  
  ADDR = 3'b000;
  
  # (tas - tes);
  
  EN_N =  1'b0;
  
  # tes;
  
  ADS_N = 1'b1;
  
  # taw;
  
  WR_N = 1'b0;
  
  # (twr - tds);
  
  dout = 8'h01;
  
  # tds;
  
  WR_N = 1'b1;
  
  # tdh;
  
  dout = 8'hz;
  
  # (twen - tdh);
  
  EN_N = 1'b1;
  
  @(negedge intr_int);
  read_ir;

end
endtask

//-----------------------------------------------------------

task cpu_write(input [2:0] address, input [7:0] din);      

	
begin
  
  ADS_N = 1'b0;
  
  # (tads - tas);
  
  ADDR = address;
  
  # (tas - tes);
  
  EN_N =  1'b0;
  
  # tes;
  
  ADS_N = 1'b1;
  
  # taw;
  
  WR_N = 1'b0;
  
  # (twr - tds);
  
  dout = din;
  
  # tds;
  
  WR_N = 1'b1;
  
  # twen;
  
  EN_N = 1'b1;
  
  # (tdh - twen);

  dout = 8'hz;
  
  
end
endtask

//-----------------------------------------------------------



task ow_write_byte(input [2:0] address, input [7:0] din);      

	
begin

  
  cpu_write(address,din);
  
  ir_en_reg = 8'h04 ^ {6'b000000,ias,1'b0};
  cpu_write(IR_ENABLE,ir_en_reg);

  @(negedge intr_int);
  read_ir;
  
  ir_en_reg = 8'h08 ^ {6'b000000,ias,1'b0};
  cpu_write(IR_ENABLE,ir_en_reg);
  
  @(negedge intr_int);
  read_ir;
end
endtask

//-----------------------------------------------------------


task ow_read_byte(input [2:0] address, output [7:0] dout);      

reg [7:0] temp;	

begin
  cpu_write(RX_BUFFER,8'hff);
  
  ir_en_reg = 8'h20 ^ {6'b000000,ias,1'b0};
  cpu_write(IR_ENABLE,ir_en_reg);
  
  
  @(negedge intr_int);
  read_ir;
  
  
  
  cpu_read(address, dout);

end
endtask


//-----------------------------------------------------------

task cpu_read(input [2:0] address, output [7:0] dout);      

	
begin
  
  ADS_N = 1'b0;
  
  # (tads - tas);
  
  ADDR = address;
  
  # (tas - tes);
  
  EN_N =  1'b0;
  
  # tes;
  
  ADS_N = 1'b1;
  
  # tar;
  
  RD_N = 1'b0;
  
  # trvd;
  
  dout = DATA;
  
  # (trd - trvd);
  
  RD_N = 1'b1;
  
  # tren;
  
  EN_N = 1'b1;
  
  # (thz - tren);
  

end
endtask

//-----------------------------------------------------------



task read_ir;
reg [7:0] ir_reg;
begin
  cpu_read(IR, ir_reg);  
end
endtask

//-----------------------------------------------------------

task set_clock_divisor;		
input [7:0] ratio;
reg [7:0] val;
begin
  $display("%t CPU      - Setting clock divisor = %0d ",$time, ratio);

   case (ratio)
      4:         val = 8'h88;
      5:         val = 8'h82;
      6:         val = 8'h85;
      7:         val = 8'h83;
      8:         val = 8'h8c;
      10:        val = 8'h86;
      12:        val = 8'h89;
      14:        val = 8'h87;
      16:        val = 8'h90;
      20:        val = 8'h8a;
      24:        val = 8'h8d;
      28:        val = 8'h8b;
      32:        val = 8'h94;
      40:        val = 8'h8e;
      48:        val = 8'h91;
      56:        val = 8'h8f;
      64:        val = 8'h98;
      80:        val = 8'h92;
      96:        val = 8'h95;
      112:       val = 8'h93;
      default: begin
                  val = 8'h00;
		  $display("%t  ERROR: Invalid Clock Divisor %d", $time, ratio);
               end
   endcase   
   
   cpu_write(CLK_DIV_REG,val);
   
   CLKSEL = ratio;
   
end
endtask

//-----------------------------------------------------------

task read_rom;		// Read ROM ////////////////////////////////////////////
begin
  $display("%t CPU      - READ_ROM command issued",$time);
  
  current_cmd = READ_ROM;
  
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(READ_ROM);

  ow_reset;
  
  
  ow_write_byte(TX_BUFFER, READ_ROM);
  
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[7:0]);
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[7:0]);
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[7:0]);
   
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[15:8]); 
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[23:16]);
  
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[31:24]);
  
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[39:32]);
  
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[47:40]);
  
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[55:48]);
  ow_read_byte(RX_BUFFER, tb_ds1wm.xtc_ds1wm.ROMID[63:56]);        // CRC-8 byte (not calculated)
//  check_valid_rom_cmd;

//  cpu_write(IR_ENABLE,8'h08);

end
endtask


//-----------------------------------------------------------

task match_rom;		// Match ROM ///////////////////////////////////////////
input [63:0] match_id;
begin
  $display("%t CPU      - MATCH_ROM command issued",$time);
  
  current_cmd = MATCH_ROM;
  
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(MATCH_ROM);
  
  ow_reset;
  ow_write_byte(TX_BUFFER, MATCH_ROM);
  
  ow_write_byte(TX_BUFFER, match_id[7:0]);
  ow_write_byte(TX_BUFFER, match_id[15:8]);
  ow_write_byte(TX_BUFFER, match_id[23:16]);
  ow_write_byte(TX_BUFFER, match_id[31:24]);
  ow_write_byte(TX_BUFFER, match_id[39:32]);
  ow_write_byte(TX_BUFFER, match_id[47:40]);
  ow_write_byte(TX_BUFFER, match_id[55:48]);
  ow_write_byte(TX_BUFFER, match_id[63:56]);
//  check_valid_rom_cmd;
end
endtask

/*
task search_rom;	// Search ROM //////////////////////////////////////////
input [63:0] search_id;
integer loop;
reg result;
begin
  $display("\nSearch rom command at time %0t",$time/div);
  current_cmd = SEARCH_ROM;
  ow_reset;
  ow_write_byte(SEARCH_ROM);
  for (loop=0; loop <= 63; loop=loop+1)
  begin
    ow_read_slot(result);  // read current id bit
    ow_read_slot(result);  // read complement of current id bit
    if (search_id[loop] == 1'b0)
      ow_write0;
    else
      ow_write1;
    end
  check_valid_rom_cmd;
end
endtask
*/

//-----------------------------------------------------------

task skip_rom;		// Skip ROM ////////////////////////////////////////////
begin
  $display("%t CPU      - SKIP_ROM command issued",$time);
  current_cmd = SKIP_ROM;
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(SKIP_ROM);
  ow_reset;
  ow_write_byte(TX_BUFFER, SKIP_ROM);
  //check_valid_rom_cmd;
end
endtask

/*
task od_match_rom;	// Overdrive Match ROM /////////////////////////////////
input [63:0] match_id;
begin
  $display("\noverdrive match command at time %0t",$time/div);
  current_cmd = OD_MATCH;
  overdrive = 0;
  ow_reset;
  ow_write_byte(OD_MATCH);
  overdrive = 1;
  ow_write_byte(match_id[7:0]);
  ow_write_byte(match_id[15:8]);
  ow_write_byte(match_id[23:16]);
  ow_write_byte(match_id[31:24]);
  ow_write_byte(match_id[39:32]);
  ow_write_byte(match_id[47:40]);
  ow_write_byte(match_id[55:48]);
  ow_write_byte(match_id[63:56]);
  check_valid_rom_cmd;
end  
endtask


task ow_reset_to_od;	// Kickdown to OD mode /////////////////////////////////
begin
  $display("\nslow down by overdrive mode reset at time %0t",$time/div);
  disable write_sp;
  disable read_sp;
  disable copy_sp;
  disable read_mem;
  disable read_mem_crc;
  disable clear_mem;
  disable force_conversion;
  disable start_mission;
  disable stop_mission;
  disable run_bist;
  disable cont_bist;
  OW_WR = 'b0;
  prot_ok = 0;		// Reset prot_ok bit
  #od_trstl;
  OW_WR = 'bz;
  #od_trsth;
  hyperdrive = 1'b0;
  overdrive = 1'b1;
end
endtask

task ow_reset_to_std;	// Kickdown to STD mode ////////////////////////////////
begin
  $display("\nslow down by standard mode reset at time %0t",$time/div);
  disable write_sp;
  disable read_sp;
  disable copy_sp;
  disable read_mem;
  disable read_mem_crc;
  disable clear_mem;
  disable force_conversion;
  disable start_mission;
  disable stop_mission;
  disable run_bist;
  disable cont_bist;
  OW_WR = 'b0;
  prot_ok = 0;		// Reset prot_ok bit
  #trstl;
  #trstl;
  OW_WR = 'bz;
  #trsth;
  hyperdrive = 1'b0;
  overdrive = 1'b0;
end
endtask

*/
///// 1922 RAM commands. ///////////////////////////////////////////////////////

//-----------------------------------------------------------

task write_sp;		// Write Scratchpad ////////////////////////////////////

begin

  $display("%t CPU      - Write_Scratchpad command issued",$time);
  current_cmd = WRITE_SP;
  
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(WRITE_SP);
  ow_write_byte(TX_BUFFER, WRITE_SP);
  
end
endtask         

//-----------------------------------------------------------

task read_sp;		// Read Scratchpad /////////////////////////////////////

begin
  $display("%t CPU      - Read_Scratchpad command issued",$time);
  current_cmd = READ_SP;
  
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(READ_SP);
//  check_protok;				// Check if ROM command succeeded.
  ow_write_byte(RX_BUFFER, READ_SP);
end
endtask 


//-----------------------------------------------------------

task copy_sp;		// Copy Scratchpad /////////////////////////////////////
      
begin
  $display("%t CPU      - Copy_Scratchpad command issued", $time);
  current_cmd = COPY_SP;
  
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(COPY_SP);
//  check_protok;				// Check if ROM command succeeded.
  ow_write_byte(TX_BUFFER, COPY_SP);
//  prot_ok = 0;				// Reset prot_ok
end
endtask


//-----------------------------------------------------------

task read_mem;		// Read Memory /////////////////////////////////////////

begin

  $display("%t CPU      - Read_Memory command issued",$time);
  current_cmd = READ_MEM;
  
  tb_ds1wm.xtc_ds1wm.xscoreboard.write_command_fifo(READ_MEM);
//  check_protok;				// Check if ROM command succeeded.
  ow_write_byte(TX_BUFFER, READ_MEM);
end
endtask

//-----------------------------------------------------------
task set_ias(input val);		

begin
   if (val == 1) 
      $display("%t CPU      - Setting Interrupt Active High",$time);
   else
      $display("%t CPU      - Setting Interrupt Active Low",$time);
     
   #1 ias = val;
end
endtask

endmodule
